/**
   Wifi_IR_Remote_TeensyLC
   https://www.forward.com.au/pfod/HomeAutomation/Wifi_IR_Remote/index.html

   (c)2017 Forward Computing and Control Pty. Ltd.
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.
*/
// IR_Send.ino
#include <IRremote.h>
IRsend irsend;

// uncomment this define to debug the LIR decoder.
// Note main .ino needs to Serial.begin()
//#define DEBUG_LIR


void  sendNEC(unsigned long data,  int nbits) {
  irsend.sendNEC(data, nbits);
}

void sendRaw(const unsigned int buf[],  unsigned int len,  unsigned int hz) {
  irsend.sendRaw(buf, len, hz);
}

// set these values from AnalysIR measurements of the marks and spaces from the actual remote
#define PANASONIC38_BITS          48
#define PANASONIC38_HDR_MARK    3440
#define PANASONIC38_HDR_SPACE   1792
// one == 1732, zero 868
#define PANASONIC38_BIT_MARK     408
#define PANASONIC38_ONE_SPACE   1320
#define PANASONIC38_ZERO_SPACE   456

// this is a copy of IRsend sendPanasonic but with freq set to 38 instead of 35
void sendPanasonic38 (unsigned int address,  unsigned long data) {
  // Set IR carrier frequency
  irsend.enableIROut(38);

  // Header
  irsend.mark(PANASONIC38_HDR_MARK);
  irsend.space(PANASONIC38_HDR_SPACE);

  // Address
  for (unsigned long  mask = 1UL << (16 - 1);  mask;  mask >>= 1) {
    irsend.mark(PANASONIC38_BIT_MARK);
    if (address & mask)  irsend.space(PANASONIC38_ONE_SPACE) ;
    else                 irsend.space(PANASONIC38_ZERO_SPACE) ;
  }

  // Data
  for (unsigned long  mask = 1UL << (32 - 1);  mask;  mask >>= 1) {
    irsend.mark(PANASONIC38_BIT_MARK);
    if (data & mask)  irsend.space(PANASONIC38_ONE_SPACE) ;
    else              irsend.space(PANASONIC38_ZERO_SPACE) ;
  }

  // Footer
  irsend.mark(PANASONIC38_BIT_MARK);
  irsend.space(0);  // Always end with the LED off
}

uint32_t getLIRpulseMark(uint8_t pairCount, char *pulsePairStart) {
  char* startMark = pulsePairStart + 10 * pairCount;
  return getLIRpulse(startMark);
}

uint32_t getLIRpulseSpace(uint8_t pairCount, char *pulsePairStart) {
  char* startSpace = pulsePairStart + 10 * pairCount + 5;
  return getLIRpulse(startSpace);
}

/*
   pulse is eather a mark or space  can get 0 for last space
   only converts the first 4 chars
   assumes there are at least 5 chars in str
*/
uint32_t getLIRpulse(char *pulse) {
  char c = pulse[4];
  uint32_t pulseLen = (uint32_t)strtol(pulse, NULL, 16);
  pulse[4] = c;
  if (pulseLen % 2 != 0) { // if odd add 65535
    pulseLen += 65535;
  }
  return pulseLen;
}

uint8_t convertHexCharToInt(char *hex) {
  uint8_t b = *hex;
  if (b >= '0' && b <= '9') {
    b = b - '0';
  }  else if (b >= 'a' && b <= 'f') {
    b = b - 'a' + 10;
  }  else if (b >= 'A' && b <= 'F') {
    b = b - 'A' + 10;
  }
  return b;
}

// assumes str starts at first pulse pair
int countLIRpulsePairs(char *str) {
  // looking for pairs of 4 hex digits stop when find just 2 chars.
  int numberOfPairs = 0;
  char *nextPair = str;
  char *firstSpace, *secondSpace;
  while (*nextPair != '\0') {
    firstSpace = strchr(nextPair, ' ');
    if (firstSpace == NULL) {
      return -1; // error either not found or not 4 chars
    }
    if ((firstSpace - nextPair) == 2) {
      // finished pairs
      break;
    }
    secondSpace = strchr(firstSpace + 1, ' ');
    if ((secondSpace == NULL) || (secondSpace - firstSpace - 1 != 4)) {
      return -2; // error no second part of pair
    }
    numberOfPairs++;
    nextPair = secondSpace + 1;
  }
  return numberOfPairs;

}

/**
   to determine if we need to send the last space we need to look ahead for a null after this seqence
   OR for 'F' null
   Note if repeating need to only drop space for last repeat
*/

bool isLastSeq(char *seqPtr) {
  if ( (*seqPtr == '\0') || // this is the end
       ( ((*seqPtr == 'F') || (*seqPtr == 'f')) && (*(seqPtr + 1) == '\0') ) // followed by just F
     )
  { // then
    return true;
  } // else
  return false;
}

/*
    Counts or fill the the number of marks and spaces for LIR code to allocate raw array for IRremote
    To just count call with rawArray == NULL
    To fill call with rawArray allocated large enough
*/
uint32_t countOrFillLIRpulses(int numberOfPairs, char* startPulsePairs, char *startSeq, unsigned int* raw) {
  char *seqPtr = startSeq;
  size_t rawIdx = 0;
  bool repeating = false;
  bool havePulse = false;
  uint8_t pulseNo = 0;
  uint8_t repeatCount = 0;


  while (*seqPtr != '\0') {
    // skip spaces
    if (*seqPtr == ' ') {
      seqPtr++;
      continue;
    }
    if ((*seqPtr == 'F') || (*seqPtr == 'f')) {
      seqPtr++;
      // start repeat
#ifdef DEBUG_LIR
      Serial.println("found repeat");
#endif
      repeating = true;
      havePulse = false;
      continue;
    }
    if (!havePulse) {
      // get pulse
      pulseNo = convertHexCharToInt(seqPtr);
      seqPtr++;
#ifdef DEBUG_LIR
      Serial.print("found pulseNo:");
      Serial.println(pulseNo);
#endif
      if (pulseNo >= numberOfPairs) {
        return -3; // invalid pulseNo
      }
      //      Serial.print(" Pulse Pair : ");
      //      Serial.print(getLIRpulseMark(pulseNo,startPulsePairs)); Serial.print(" "); Serial.println(getLIRpulseSpace(pulseNo,startPulsePairs));

      if (!repeating) {

        //check for dropSpace
        bool dropSpace = isLastSeq(seqPtr);
        if (dropSpace) {
#ifdef DEBUG_LIR
          Serial.println(" End dropSpace");
#endif
        }
        // send mark
        uint32_t mark = getLIRpulseMark(pulseNo, startPulsePairs);
#ifdef DEBUG_LIR
        Serial.print(mark); Serial.print(" ");
#endif
        if (raw != NULL) {
          raw[rawIdx++] = mark;
        } else {
          rawIdx++;
        }
        if (!dropSpace) {
          // send space
          uint32_t space = getLIRpulseSpace(pulseNo, startPulsePairs);
          if (raw != NULL) {
            raw[rawIdx++] = space;
          } else {
            rawIdx++;
          }
#ifdef DEBUG_LIR
          Serial.print(space); Serial.println(" ");
#endif
        } else {
#ifdef DEBUG_LIR
          Serial.println();
#endif
        }
        // finished this single pulse
        repeating = false;
        repeatCount = 0;
        havePulse = false;

      } else {
        // repeating
        havePulse = true;
        continue;
      }
    } else {  // have pulse should be a repeat
      if (!repeating) {
#ifdef DEBUG_LIR
        Serial.println("Error should be repeating here");
#endif
      }
      repeatCount = convertHexCharToInt(seqPtr);
      seqPtr++;
#ifdef DEBUG_LIR
      Serial.print("found repeatCount:");
      Serial.println(repeatCount);
#endif
      bool dropSpace = isLastSeq(seqPtr);
      if (dropSpace) {
#ifdef DEBUG_LIR
        Serial.println(" End dropSpace last");
#endif
      }

      for (uint8_t i = 0; i < repeatCount; i++) {
        // send mark
        uint32_t mark = getLIRpulseMark(pulseNo, startPulsePairs);
        if (raw != NULL) {
          raw[rawIdx++] = mark;
        } else {
          rawIdx++;
        }
#ifdef DEBUG_LIR
        Serial.print(mark); Serial.print(" ");
#endif
        if (dropSpace && (i == repeatCount - 1)) {
          // skip last space
#ifdef DEBUG_LIR
          Serial.println();
#endif
        } else {
          // send space
          uint32_t space = getLIRpulseSpace(pulseNo, startPulsePairs);
          //     irsend.space(space);
          if (raw != NULL) {
            raw[rawIdx++] = space;
          } else {
            rawIdx++;
          }
#ifdef DEBUG_LIR
          Serial.print(space); Serial.println(" ");
#endif
        }
      }
      // finish repeat
      repeatCount = 0;
      repeating = false;
      havePulse = false;
    }
  }
  return rawIdx;
}

// returns number of raw code or -ve number for errors
int parseAndSendLIRcode(char *codeStr) {
  unsigned int* raw = NULL;
  //size_t rawIdx = 0;
  char* code = codeStr;
  const char LIR[] = "LIR: ";
  if (strstr(code, LIR) != code) {
    // error not an LIR: code
    return -1;
  }
  code = code + strlen(LIR); // skip LIR:
  // pick up freq
  char *nextSpace = strchr(code, ' ');
  if (nextSpace == NULL) {
    // code too short
    return -2;
  }
  // else
  *nextSpace = '\0'; // terminate there
  long freq = strtol(code, NULL, 16);
#ifdef DEBUG_LIR
  Serial.print("Freq:"); Serial.println(freq);
#endif

  // irsend.enableIROut(freq); // set freq in khZ

  *nextSpace = ' '; // restore space for next time
  char* startPulsePairs = nextSpace + 1;
  int numberOfPairs = countLIRpulsePairs(startPulsePairs);
  if (numberOfPairs <= 0) {
    return -10 + numberOfPairs;
  }
#ifdef DEBUG_LIR
  Serial.print("No of Pairs : "); Serial.println(numberOfPairs);
  // print out pulses
  for (int i = 0; i < numberOfPairs; i++) {
    Serial.print(getLIRpulseMark(i, startPulsePairs)); Serial.print(" "); Serial.println(getLIRpulseSpace(i, startPulsePairs));
  }
#endif

  // ok now parse sequence and send
  char *seqPtr = startPulsePairs + numberOfPairs * 10;
  size_t rawLen = countOrFillLIRpulses(numberOfPairs, startPulsePairs, seqPtr, NULL);
  if (rawLen <= 0 ) {
    return -20;
  }
  raw = (unsigned int*)malloc(rawLen * sizeof(unsigned int));
  if (raw == NULL) {
    return -30; // not enough memory
  }
  rawLen = countOrFillLIRpulses(numberOfPairs, startPulsePairs, seqPtr, raw);
  if (rawLen <= 0 ) {
    return -20;
  }

#ifdef DEBUG_LIR
  Serial.print(" Freq (kHz):"); Serial.println(freq);
  Serial.print("Raw ("); Serial.print(rawLen); Serial.println(") :");
  for (size_t i = 0; i < rawLen; i++) {
    Serial.print(", "); Serial.print(raw[i]);
  }
  Serial.println();
#endif
  irsend.sendRaw(raw, rawLen, freq);
  free(raw); // free the memory
  return 0; // OK
}


uint8_t reverseByte(uint8_t b) {
  b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
  b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
  b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
  return b;
}

/**
   Replaces bytes in the array with the reversed bits in each byte
*/
uint8_t* reverse8_byteArray(uint8_t* bytePtr, size_t len) {
  for (size_t i = 0;  i < len ; i++) {
    bytePtr[i] = reverseByte(bytePtr[i]);
  }
  return bytePtr;
}

/**
   Replaces bytes in the uint16_t number (2 bytes) with the reversed bits in each byte
*/
uint16_t reverse8_uint16(uint16_t in) {
  uint16_t out = in;
  reverse8_byteArray((uint8_t*)&out, sizeof(uint16_t));
  return out;
}

/**
   Replaces bytes in the uint32_t number (4 bytes) with the reversed bits in each byte
*/
uint32_t reverse8_uint32(uint32_t in) {
  uint32_t out = in;
  reverse8_byteArray((uint8_t*)&out, sizeof(uint32_t));
  return out;
}

/**
   Replaces bytes in the uint64_t number (8 bytes) with the reversed bits in each byte
*/
uint64_t reverse8_uint64(uint64_t in) {
  uint64_t out = in;
  reverse8_byteArray((uint8_t*)&out, sizeof(uint64_t));
  return out;
}


// from https://stackoverflow.com/questions/2182002/convert-big-endian-to-little-endian-in-c-without-using-provided-func
/**
   Converts uint16_t (2 bytes) from MSB to LSB and back again
*/
uint16_t swap_uint16( uint16_t val ) {
  return (val << 8) | (val >> 8 );
}

/**
   Converts uint32_t (4 bytes) from MSB to LSB and back again
*/
uint32_t swap_uint32( uint32_t val ) {
  val = ((val << 8) & 0xFF00FF00 ) | ((val >> 8) & 0xFF00FF );
  return (val << 16) | (val >> 16);
}

/**
   Converts uint64_t (8 bytes) from MSB to LSB and back again
*/
uint64_t swap_uint64( uint64_t val ) {
  val = ((val << 8) & 0xFF00FF00FF00FF00ULL ) | ((val >> 8) & 0x00FF00FF00FF00FFULL );
  val = ((val << 16) & 0xFFFF0000FFFF0000ULL ) | ((val >> 16) & 0x0000FFFF0000FFFFULL );
  return (val << 32) | (val >> 32);
}
